@@ -8,26 +8,26 @@ import functools
 import logging
 import time
 import math
+import qrcode
+from io import BytesIO
+import re
 
 import gi
 gi.require_version('Gtk', '3.0')
 gi.require_version('XApp', '1.0')
-from gi.repository import Gtk, GLib, XApp, Gio, GObject, Gdk
+from gi.repository import Gtk, GLib, XApp, Gio, GObject, Gdk, GdkPixbuf
 
 import config
 import prefs
 import util
+import dbus_service
 import server
 import auth
+import misc
 import networkmonitor
 from ops import SendOp, ReceiveOp
 from util import TransferDirection, OpStatus, RemoteStatus
 
-# Don't let warp run as root
-if os.getuid() == 0:
-    print("Warpinator should not be run as root. Please run it in user mode.")
-    sys.exit(1)
-
 # XApp 2.0 required for favorites.
 HAVE_XAPP_FAVORITES = True
 
@@ -45,7 +45,7 @@ _ = gettext.gettext
 setproctitle.setproctitle("warpinator")
 
 SERVER_RESTART_TIMEOUT = 15
-SERVER_START_TIMEOUT = 8
+SERVER_START_TIMEOUT = 10
 DISCOVERY_TIMEOUT = 3
 
 SECURE_MODE_EXIT_TIME_SECONDS = 60 * 60 # 60 minutes in seconds
@@ -151,12 +151,12 @@ class OpItem(object):
                 self.op_transfer_status_message.set_text(_("Waiting for approval"))
             else:
                 self.op_transfer_status_message.set_text(_("Waiting for your approval"))
-                if self.op.existing and prefs.prevent_overwriting():
-                    self.op_transfer_problem_label.show()
-                    self.op_transfer_problem_label.set_text(_("Files may be overwritten"))
-                elif not self.op.have_space:
+                if not self.op.have_space:
                     self.op_transfer_problem_label.show()
                     self.op_transfer_problem_label.set_text(_("Not enough disk space"))
+                elif self.op.existing and prefs.prevent_overwriting():
+                    self.op_transfer_problem_label.show()
+                    self.op_transfer_problem_label.set_text(_("Files may be overwritten"))
         elif (self.op.status == OpStatus.CANCELLED_PERMISSION_BY_SENDER and isinstance(self.op, SendOp)) or \
             (self.op.status == OpStatus.CANCELLED_PERMISSION_BY_RECEIVER and isinstance(self.op, ReceiveOp)):
             self.op_transfer_status_message.set_text(_("Request cancelled"))
@@ -176,7 +176,7 @@ class OpItem(object):
         elif self.op.status == OpStatus.FILE_NOT_FOUND:
             self.op_transfer_problem_label.show()
             self.op_transfer_status_message.set_text(_("Transfer failed"))
-            if self.op.first_missing_file != None:
+            if self.op.first_missing_file is not None:
                 self.op_transfer_problem_label.set_text(_("File '%s' not found") % self.op.first_missing_file)
             else:
                 self.op_transfer_problem_label.set_text(_("Some files not found"))
@@ -485,6 +485,14 @@ class WarpWindow(GObject.Object):
         self.secmo_infobar.set_revealed(False)
         self.secmo_infobar_prefs_button.connect("clicked", self.open_prefs_networking)
 
+        self.sandbox_mode_infobar = self.builder.get_object("sandbox_mode_infobar")
+        self.update_sandbox_mode_infobar()
+
+        self.sandbox_mode_infobar_close_button = self.builder.get_object("sandbox_infobar_close_button")
+        self.sandbox_mode_infobar_close_button.connect("clicked", lambda b: self.sandbox_mode_infobar.hide())
+        self.sandbox_mode_more_info_button = self.builder.get_object("sandbox_infobar_more_info_button")
+        self.sandbox_mode_more_info_button.connect("clicked", self.sandbox_info_callback)
+
         # user view
         self.user_back_button = self.builder.get_object("user_back")
         self.user_back_button.connect("clicked", self.back_to_overview)
@@ -503,6 +511,22 @@ class WarpWindow(GObject.Object):
         self.user_online_spinner = self.builder.get_object("user_online_spinner")
         self.user_clear_ops_button = self.builder.get_object("user_clear_ops_button")
 
+        self.app_restart_ops_count_label = self.builder.get_object("app_restart_ops_count_label")
+        self.app_restart_spinner = self.builder.get_object("app_restart_spinner")
+        self.app_restart_icon = self.builder.get_object("app_restart_icon")
+        self.app_restart_button = self.builder.get_object("app_restart_button")
+        self.app_restart_button.connect("clicked", self.app_restart_button_clicked)
+
+        self.no_disk_space_label = self.builder.get_object("no_disk_space_label")
+        self.no_disk_space_trash_button = self.builder.get_object("no_disk_space_trash_button")
+        self.no_disk_space_trash_button.connect("clicked", self.no_disk_space_trash_clicked)
+        self.no_disk_space_trash_button.set_visible(util.trash_uri_supported())
+        self.no_disk_space_open_save_button = self.builder.get_object("no_disk_space_open_save_button")
+        self.no_disk_space_open_save_button.connect("clicked", self.no_disk_space_open_save_clicked)
+        self.no_disk_space_baobab_button = self.builder.get_object("no_disk_space_baobab_button")
+        self.no_disk_space_baobab_button.connect("clicked", self.no_disk_space_baobab_clicked)
+        self.no_disk_space_baobab_button.set_visible(util.disk_usage_available())
+
         # Send Files button
         main_menu = Gtk.Menu()
 
@@ -510,12 +534,12 @@ class WarpWindow(GObject.Object):
         # user's desktop and flatpaks.
         if (not config.FLATPAK_BUILD) and HAVE_XAPP_FAVORITES:
             if XApp.Favorites.get_default().get_n_favorites() > 0:
-                item = Gtk.MenuItem(_("Favorites"))
+                item = Gtk.MenuItem(label=_("Favorites"))
                 main_menu.add(item)
                 favorites = XApp.Favorites.get_default().create_menu(None, self.favorite_selected)
                 item.set_submenu(favorites)
 
-        item = Gtk.MenuItem(_("Recents"))
+        item = Gtk.MenuItem(label=_("Recents"))
         main_menu.add(item)
         self.recent_menu = util.get_recent_chooser_menu()
         self.recent_menu.connect("item-activated", self.recent_item_selected)
@@ -541,6 +565,11 @@ class WarpWindow(GObject.Object):
         item.connect("activate", self.on_open_location_clicked)
         menu.add(item)
 
+        item = Gtk.MenuItem(label=_("Connect manually"), sensitive=False)
+        item.connect("activate", self.manual_connect)
+        menu.add(item)
+        self.manual_connect_menu_item = item
+
         item = Gtk.MenuItem(label=_("Preferences"))
         item.connect("activate", self.open_preferences)
         menu.add(item)
@@ -577,6 +606,12 @@ class WarpWindow(GObject.Object):
     def show_about(self, widget):
         util.AboutDialog(self.window)
 
+    def sandbox_info_callback(self, widget):
+        try:
+            Gtk.show_uri_on_window(self.window, "https://github.com/linuxmint/warp/blob/master/README.md#folder-isolation", Gtk.get_current_event_time())
+        except GLib.Error as e:
+            logging.warn("Couldn't open readme url: %s" % e.message)
+
     def window_delete_event(self, widget, event, data=None):
         if prefs.use_tray_icon():
             self.window.hide()
@@ -594,6 +629,13 @@ class WarpWindow(GObject.Object):
 
         return Gdk.EVENT_PROPAGATE
 
+    def show_page(self, page):
+        # If we need to restart, don't let anything interrupt it.
+        if self.view_stack.get_visible_child_name() == "restart":
+            return
+
+        self.view_stack.set_visible_child_name(page)
+
     def toggle_visibility(self, time=0):
         if self.window.is_active():
             self.window.hide()
@@ -636,10 +678,13 @@ class WarpWindow(GObject.Object):
         timeout = SERVER_RESTART_TIMEOUT if restarting else SERVER_START_TIMEOUT
 
         self.server_start_timeout_id = GLib.timeout_add_seconds(timeout, self.server_not_started_timeout)
-        self.view_stack.set_visible_child_name("startup")
+        self.manual_connect_menu_item.set_sensitive(False)
+
+        self.update_sandbox_mode_infobar()
+        self.show_page("startup")
 
     def server_not_started_timeout(self):
-        self.view_stack.set_visible_child_name("server-problem")
+        self.show_page("server-problem")
 
         if not self.netmon.online:
             self.something_wrong_label.set_text(_("You don't appear to be connected to a network."))
@@ -649,6 +694,10 @@ class WarpWindow(GObject.Object):
         self.server_start_timeout_id = 0
         return False
 
+    def update_sandbox_mode_infobar(self):
+        insecure = (config.sandbox_mode == "legacy") and (util.home_is_writable() if config.FLATPAK_BUILD else True)
+        self.sandbox_mode_infobar.set_visible(insecure)
+
     def show_no_network(self):
         if self.server_start_timeout_id > 0:
             GLib.source_remove(self.server_start_timeout_id)
@@ -714,17 +763,21 @@ class WarpWindow(GObject.Object):
         Gtk.drag_finish(context, True, False, _time)
         self.drop_pending = False
 
-    def update_local_user_info(self, ip="0.0.0.0", iface=""):
+    def update_local_user_info(self, ip="0.0.0.0", iface="", auth_port=0):
         self.app_local_name_label.set_text(util.get_local_name())
         self.app_iface_label.set_text(iface)
         self.app_ip_label.set_text(ip)
+        self.current_ip = ip
+        self.current_iface = iface
+        self.current_auth_port = auth_port
 
     def update_secure_mode_info(self, is_secure):
         if is_secure:
             self.stop_secure_mode_timer()
             self.secmo_infobar.set_revealed(False)
-            self.secmo_infobar.hide()
+            self.secmo_infobar.hide() # shouldn't be needed but in some themes, the bar isn't fully hidden when revealed=false
         else:
+            self.secmo_infobar.set_revealed(True)
             self.start_secure_mode_timer()
 
     def start_secure_mode_timer(self):
@@ -764,15 +817,47 @@ class WarpWindow(GObject.Object):
     def on_open_location_clicked(self, widget, data=None):
         util.open_save_folder()
 
-    def open_prefs_networking(self, button):
-        prefs.Preferences(self.window, "network")
+    def open_prefs_networking(self, button, data=None):
+        self.open_preferences(None, "network")
 
-    def open_preferences(self, menuitem, data=None):
-        prefs.Preferences(self.window, "general")
+    def open_preferences(self, menuitem, data="general"):
+        self.prefs_window = prefs.Preferences(self.window, data)
+        self.prefs_window.window.connect("destroy", self.on_prefs_destroy)
+
+    def on_prefs_destroy(self, window):
+        self.prefs_window = None
+
+    def manual_connect(self, _):
+        self.connect_dialog = ManualConnectDialog(self)
+        self.connect_dialog.show()
+
+    @GObject.Signal(arg_types=(str,))
+    def manual_connect_to_host(self, host):
+        logging.debug("Connecting to " + host)
 
     def report_bad_save_folder(self):
-        self.bad_save_folder_label.set_text(prefs.get_save_path())
-        self.view_stack.set_visible_child_name("bad-save-folder")
+        path = prefs.get_save_path()
+        if path.startswith("/run/user"):
+            path = os.path.basename(path)
+        self.bad_save_folder_label.set_text(path)
+        self.show_page("bad-save-folder")
+
+    def report_no_disk_space(self):
+        self.no_disk_space_label.set_label(
+            _("Not enough disk space remaining to receive files (%s is reserved).") % \
+                GLib.format_size(prefs.get_min_free_space() * 1024 * 1024)
+        )
+
+        self.show_page("no-disk-space")
+
+    def no_disk_space_trash_clicked(self, button):
+        util.open_trash()
+
+    def no_disk_space_open_save_clicked(self, button):
+        util.open_save_folder()
+
+    def no_disk_space_baobab_clicked(self, button):
+        util.open_disk_usage()
 
     def add_remote_button(self, remote_machine, simulated=False):
         if self.discovery_time_out_id > 0:
@@ -793,7 +878,7 @@ class WarpWindow(GObject.Object):
 
         self.user_list_box.add(button.button)
 
-        self.view_stack.set_visible_child_name("overview")
+        self.show_page("overview")
         self.sort_buttons()
 
     def remove_remote_button(self, remote_machine):
@@ -816,7 +901,7 @@ class WarpWindow(GObject.Object):
             self.user_list_box.remove(child)
 
     def focus_remote_machine(self, remote_machine, data=None):
-        if self.current_selected_remote_machine != None:
+        if self.current_selected_remote_machine is not None:
             self.cleanup_user_view()
 
         self.switch_to_user_view(remote_machine)
@@ -825,7 +910,39 @@ class WarpWindow(GObject.Object):
             self.toggle_visibility()
 
     def display_shutdown(self):
-        self.view_stack.set_visible_child_name("shutdown")
+        self.manual_connect_menu_item.set_sensitive(False)
+        self.show_page("shutdown")
+
+    def display_restart(self):
+        if self.prefs_window is not None:
+            self.prefs_window.destroy()
+            self.prefs_window = None
+
+        self.show_page("restart")
+
+    def update_restart_dialog_status(self, active_ops):
+        if active_ops > 0:
+            label = gettext.ngettext(
+                _("Waiting for %d operation to complete"),
+                _("Waiting for %d operations to complete"),
+                active_ops
+            ) % active_ops
+
+            self.app_restart_ops_count_label.set_label(label)
+            self.app_restart_ops_count_label.show()
+            self.app_restart_spinner.show()
+            self.app_restart_icon.hide()
+            self.app_restart_button.set_sensitive(False)
+        else:
+            self.app_restart_ops_count_label.set_label("")
+            self.app_restart_spinner.hide()
+            self.app_restart_icon.show()
+            self.app_restart_button.set_sensitive(True)
+            self.app_restart_button.get_style_context().add_class("suggested-action")
+
+    def app_restart_button_clicked(self, button, data=None):
+        self.display_shutdown()
+        self.emit("exit")
 
     def notify_server_started(self):
         self.server_restarting = False
@@ -849,13 +966,14 @@ class WarpWindow(GObject.Object):
         if self.server_restarting:
             return
 
+        self.manual_connect_menu_item.set_sensitive(True)
         self.discovery_time_out_id = GLib.timeout_add_seconds(DISCOVERY_TIMEOUT, self.discovery_timed_out)
-        self.view_stack.set_visible_child_name("discovery")
+        self.show_page("discovery")
 
     def discovery_timed_out(self):
         logging.debug("UI: Discovery timed out (no remotes)")
 
-        self.view_stack.set_visible_child_name("no-remotes")
+        self.show_page("no-remotes")
         self.discovery_time_out_id = 0
         return False
 
@@ -872,7 +990,7 @@ class WarpWindow(GObject.Object):
             self.window.set_urgency_hint(True)
 
     def switch_to_user_view(self, remote_machine):
-        self.view_stack.set_visible_child_name("user")
+        self.show_page("user")
         self.current_selected_remote_machine = remote_machine
 
         self.refresh_remote_machine_view()
@@ -904,7 +1022,7 @@ class WarpWindow(GObject.Object):
 
         self.user_ip_label.set_text(str(remote.ip_info.ip4_address))
 
-        if remote.avatar_surface != None:
+        if remote.avatar_surface is not None:
             self.user_avatar_image.set_from_surface(remote.avatar_surface)
         else:
             self.user_avatar_image.set_from_icon_name("avatar-default-symbolic", Gtk.IconSize.DND)
@@ -984,7 +1102,7 @@ class WarpWindow(GObject.Object):
 
     def back_to_overview(self, button=None, data=None):
         if not self.server_restarting:
-            self.view_stack.set_visible_child_name("overview")
+            self.show_page("overview")
 
         self.cleanup_user_view()
 
@@ -993,7 +1111,7 @@ class WarpWindow(GObject.Object):
         # we just visited.
         buttons = self.user_list_box.get_children()
 
-        if self.current_selected_remote_machine != None:
+        if self.current_selected_remote_machine is not None:
             self.current_selected_remote_machine.stamp_recent_time()
 
             for child in buttons:
@@ -1039,6 +1157,93 @@ class WarpWindow(GObject.Object):
     def destroy(self):
         self.window.destroy()
 
+class ManualConnectDialog(Gtk.Window):
+    def __init__(self, parent:WarpWindow):
+        super().__init__(title=_("Manual connection"), transient_for=parent.window, modal=True, resizable=False)
+        self.parent = parent
+
+        self.ip_validator_re = re.compile(r'(warpinator://)?(\d{1,3}(\.\d{1,3}){3}):(\d{1,6})/?$')
+        self.connecting = False
+
+        self.set_default_size(150, 100)
+        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
+        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
+
+        self.builder = Gtk.Builder.new_from_file(os.path.join(config.pkgdatadir, "manual-connect.ui"))
+        top_box = self.builder.get_object("top_box")
+        self.add(top_box)
+
+        self.entry = self.builder.get_object("ip_entry")
+        self.connect_button = self.builder.get_object("connect_button")
+        self.status_label = self.builder.get_object("status_label")
+        ip_label = self.builder.get_object("our_ip_label")
+        qr_holder = self.builder.get_object("qr_holder")
+
+        self.entry.connect("changed", self.validate_address)
+        self.connect_button.connect("clicked", self.on_connecting)
+
+        # Target size is 360 total * ui scale
+        # version 1 is 21x21 blocks
+        # border is also in blocks.
+        # box_size is pixels-per-block
+        #
+        # qr code pixels width is (21 + (2 * border)) * box_size
+        # 24 (total block width) * 15 (box_size) = 360
+
+        qrbytes = BytesIO()
+        qr = qrcode.QRCode(
+            version=1,
+            box_size=15 * self.get_scale_factor(),
+            border=2
+        )
+
+        qr.add_data("warpinator://%s:%d" % (parent.current_ip, parent.current_auth_port))
+        img = qr.make_image()
+        img.save(qrbytes, "BMP")
+
+        stream = Gio.MemoryInputStream.new_from_bytes(GLib.Bytes.new(qrbytes.getvalue()))
+        pixbuf = GdkPixbuf.Pixbuf.new_from_stream_at_scale(stream, 360 * self.get_scale_factor(), -1, True, None)
+        surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, self.get_scale_factor(), None)
+        qr_image = Gtk.Image.new_from_surface(surface)
+        qr_holder.add(qr_image)
+
+        ip_label.set_label("%s:%d" % (parent.current_ip, parent.current_auth_port))
+
+        self.set_focus(qr_holder)
+        self.show_all()
+
+    def on_connecting(self, _btn):
+        self.connecting = True
+
+        self.connect_button.set_sensitive(False)
+        def pulse():
+            self.entry.progress_pulse()
+            if self.connecting:
+                return GLib.SOURCE_CONTINUE
+
+            self.entry.set_progress_fraction(0.0)
+            return GLib.SOURCE_REMOVE
+
+        GLib.timeout_add(100, pulse)
+
+        self.status_label.set_label(_("Connecting..."))
+        host = self.entry.get_text()
+        self.parent.emit("manual_connect_to_host", host)
+
+    def on_connection_result(self, result, msg):
+        self.connecting = False
+        self.connect_button.set_sensitive(True)
+
+        if result:
+            self.close()
+        else:
+            self.status_label.set_label("<span color='red'>%s</span>" % msg)
+
+    def validate_address(self, entry):
+        address = entry.get_text()
+        m = self.ip_validator_re.match(address)
+        self.connect_button.set_sensitive(m is not None)
+
 class WarpApplication(Gtk.Application):
     def __init__(self, testing=False):
         super(WarpApplication, self).__init__(application_id="org.x.Warpinator", register_session=True)
@@ -1050,13 +1255,15 @@ class WarpApplication(Gtk.Application):
         self.test_mode = testing
 
         self.bad_folder = False
-        self.save_folder_monitor = None
 
         self.inhibit_count = 0
         self.inhibit_cookie = 0
 
+        self.save_folder_poll_timer_id = 0
+        self.app_restarting = False
         self.netmon = None
         self.server = None
+        self.dbus_service = dbus_service.WarpinatorDBusService()
 
         # Set if the network state or somethign else changes while the server is starting,
         # restart it as soon as possible.
@@ -1066,6 +1273,35 @@ class WarpApplication(Gtk.Application):
         self.current_auth_port = None
         self.current_ip_info = None
 
+    def do_dbus_register(self, connection, path):
+        self.dbus_service.register(connection, path)
+        self.dbus_service.connect("handle-get-live-remotes", self.handle_dbus_get_live_remotes)
+        self.dbus_service.connect("handle-send-files", self.handle_dbus_send_files)
+
+        return Gio.Application.do_dbus_register(self, connection, path)
+
+    def do_dbus_unregister(self, connection, path):
+        self.dbus_service.unregister(connection, path)
+        Gio.Application.do_dbus_unregister(self, connection, path)
+
+    def handle_dbus_get_live_remotes(self, service):
+        if self.server is None:
+            return []
+
+        all_remotes = self.server.list_remote_machines()
+        online_remotes = [remote for remote in all_remotes if remote.status == RemoteStatus.ONLINE]
+        return online_remotes
+
+    def handle_dbus_send_files(self, service, remote_ident, files):
+        if self.server is None:
+            return
+
+        remotes = self.server.list_remote_machines()
+        for remote in remotes:
+            if remote.ident == remote_ident:
+                remote.send_files(files, dbus_sent=True)
+                break
+
     def do_startup(self):
         Gtk.Application.do_startup(self)
         logging.info("Initializing Warpinator")
@@ -1076,10 +1312,16 @@ class WarpApplication(Gtk.Application):
         action = Gio.SimpleAction.new("notification-response", vt)
         self.add_action(action)
 
+        self.secure_mode_enforcer = prefs.SecureModePrefsBlocker()
+        self.secure_mode_enforcer.start_monitor()
+        util.initialize_free_space_monitor()
+        util.free_space_monitor.connect("low-space", self.handle_low_disk_space)
+        util.free_space_monitor.connect("folder-changed", self.on_receiving_folder_changed)
+
     def do_activate(self):
         Gtk.Application.do_activate(self)
 
-        if self.window != None:
+        if self.window is not None:
             self.window.show()
             return
 
@@ -1087,6 +1329,7 @@ class WarpApplication(Gtk.Application):
 
         self.window = WarpWindow()
         self.window.connect("exit", lambda w: self.exit_warp())
+        self.window.connect("manual_connect_to_host", lambda _, host: self.server.register_with_host(host))
 
         self.add_window(self.window.window)
 
@@ -1096,39 +1339,66 @@ class WarpApplication(Gtk.Application):
 
         self.update_status_icon_from_preferences()
 
-        GLib.timeout_add(200, self.check_save_folder)
-
-    def check_save_folder(self, data=None):
-        if not util.verify_save_folder():
-            if not self.window.window.get_visible():
-                self.window.window.show()
-                self.window.window.present_with_time(Gtk.get_current_event_time())
-
-            self.window.report_bad_save_folder()
-            return GLib.SOURCE_CONTINUE
-
-        GLib.idle_add(self.start_network_monitor)
-        return GLib.SOURCE_REMOVE
-
-    def start_network_monitor(self):
         self.netmon = networkmonitor.get_network_monitor()
         self.netmon.start()
         self.netmon.connect("state-changed", self.network_state_changed)
-
         self.new_server()
 
     def network_state_changed(self, netmon, online):
         self.new_server()
 
+    def start_save_folder_check(self, server=None):
+        self.stop_save_folder_check()
+        self.save_folder_poll_timer_id = GLib.timeout_add(1000, self.check_save_folder)
+
+    def stop_save_folder_check(self):
+        if self.save_folder_poll_timer_id > 0:
+            GLib.source_remove(self.save_folder_poll_timer_id)
+            self.save_folder_poll_timer_id = 0
+
+    def check_save_folder(self, data=None):
+        perms_ok = util.verify_save_folder()
+        space_ok = util.free_space_monitor.have_enough_free(0)
+
+        if not (perms_ok and space_ok):
+            if not self.bad_folder and not self.window.window.get_visible():
+                self.window.window.show()
+                self.window.window.present_with_time(Gtk.get_current_event_time())
+                self.bad_folder = True
+            if not perms_ok:
+                self.window.report_bad_save_folder()
+            elif not space_ok:
+                self.window.report_no_disk_space()
+
+            return GLib.SOURCE_CONTINUE
+
+        self.bad_folder = False
+        GLib.idle_add(self.new_server_continue)
+        self.save_folder_poll_timer_id = 0
+        return GLib.SOURCE_REMOVE
+
     def new_server(self):
+        if self.app_restarting:
+            logging.debug("Trying to start server while there's a pending app restart. Ignoring.")
+            return
+
         if self.server_starting:
             logging.debug("Trying to start server while server already starting, will check later")
             self.server_state_dirty = True
             return
 
-        self.server_state_dirty = False
         self.server_starting = True
 
+        if self.server:
+            self.window.start_startup_timer(True)
+            self.server.connect("shutdown-complete", self.start_save_folder_check)
+            self.server.shutdown()
+        else:
+            self.start_save_folder_check()
+
+    def new_server_continue(self):
+        self.server_state_dirty = False
+
         self.window.start_startup_timer(restarting=False)
 
         self.current_port = prefs.get_port()
@@ -1137,53 +1407,45 @@ class WarpApplication(Gtk.Application):
 
         logging.debug("New server requested for '%s' (%s)", self.current_ip_info.iface, self.current_ip_info.ip4_address)
 
-        self.window.update_local_user_info(self.current_ip_info.ip4_address, self.current_ip_info.iface)
+        self.window.update_local_user_info(self.current_ip_info.ip4_address, self.current_ip_info.iface, self.current_auth_port)
 
         self.window.clear_remotes()
 
-        def try_to_start(srv=None):
-            try:
-                self.server.disconnect_by_func(try_to_start)
-            except:
-                pass
+        try:
+            self.server.disconnect_by_func(try_to_start)
+        except:
+            pass
 
-            self.server = None
+        self.server = None
 
-            auth_singleton = auth.get_singleton()
-            try:
-                auth_singleton.disconnect_by_func(self.on_group_code_changed)
-            except:
-                pass
+        auth_singleton = auth.get_singleton()
+        try:
+            prefs.prefs_settings.disconnect_by_func(self.on_group_code_changed)
+        except:
+            pass
 
-            auth_singleton.update(self.current_ip_info, self.current_port)
-            auth_singleton.connect("group-code-changed", self.on_group_code_changed)
-            
-            if auth.get_secure_mode():
-                self.window.update_secure_mode_info(True)
-            else:
-                logging.warn("Secure mode not enabled, restricting preferences.")
-                logging.warn("-- See https://github.com/linuxmint/warpinator/blob/master/README.md")
-                self.window.update_secure_mode_info(False)
-            prefs.secure_mode_blocker.enforce_secure_mode()
-
-            if not self.netmon.online:
-                logging.info("No network access")
-                self.server_starting = False
-                self.window.show_no_network()
-                return
-
-            self.server = server.Server(self.current_ip_info, self.current_port, self.current_auth_port)
-            self.server.connect("server-started", self._server_started)
-            self.server.connect("remote-machine-added", self._remote_added)
-            self.server.connect("remote-machine-removed", self._remote_removed)
-            self.server.connect("remote-machine-ops-changed", self._remote_ops_changed)
-
-        if self.server:
-            self.window.start_startup_timer(True)
-            self.server.connect("shutdown-complete", try_to_start)
-            self.server.shutdown()
+        auth_singleton.update(self.current_ip_info, self.current_port)
+        prefs.prefs_settings.connect("changed::group-code", self.on_group_code_changed)
+        
+        if prefs.get_secure_mode():
+            self.window.update_secure_mode_info(True)
         else:
-            try_to_start();
+            logging.warn("Secure mode not enabled, restricting preferences.")
+            logging.warn("-- See https://github.com/linuxmint/warpinator/blob/master/README.md")
+            self.window.update_secure_mode_info(False)
+
+        if not self.netmon.online:
+            logging.info("No network access")
+            self.server_starting = False
+            self.window.show_no_network()
+            return
+
+        self.server = server.Server(self.current_ip_info, self.current_port, self.current_auth_port)
+        self.server.connect("server-started", self._server_started)
+        self.server.connect("remote-machine-added", self._remote_added)
+        self.server.connect("remote-machine-removed", self._remote_removed)
+        self.server.connect("remote-machine-ops-changed", self._remote_ops_changed)
+        self.server.connect("manual-connect-result", self._manual_connect_result)
 
     def _server_started(self, local_machine):
         self.server_starting = False
@@ -1207,6 +1469,9 @@ class WarpApplication(Gtk.Application):
         if self.netmon:
             self.netmon.stop()
 
+        util.free_space_monitor.shutdown()
+        self.stop_save_folder_check()
+
         if self.server:
             self.setup_kill_as_a_last_resort()
             self.server.shutdown()
@@ -1222,12 +1487,13 @@ class WarpApplication(Gtk.Application):
         self.window.destroy()
 
         logging.debug("Shutdown complete")
+
         Gio.Application.do_shutdown(self)
 
     def exit_warp(self):
         GLib.idle_add(self.quit)
 
-    @util._async
+    @misc._async
     def setup_kill_as_a_last_resort(self):
         # There are plenty of opportunities for our threads to hang due to network
         # hangs and grpc errors.  Give 10 seconds and then just end it regardless.
@@ -1241,9 +1507,24 @@ class WarpApplication(Gtk.Application):
     def firewall_script_finished(self):
         self.new_server()
 
-    def on_group_code_changed(self, auth_manager):
+    def on_group_code_changed(self, settings, key):
         self.new_server()
 
+    def on_receiving_folder_changed(self, monitor):
+        if config.sandbox_mode != "bubblewrap":
+            self.new_server()
+            return
+
+        self.app_restarting = True
+        self.window.display_restart()
+        active_ops = self.server.get_active_op_count() if self.server else 0
+        self.window.update_restart_dialog_status(active_ops)
+
+    def handle_low_disk_space(self, monitor):
+        if self.server is not None and not self.server_starting:
+            self.server.cancel_all_ops()
+            self.new_server()
+
     def _remote_added(self, local_machine, remote_machine):
         self.window.add_remote_button(remote_machine)
         remote_machine.connect("machine-info-changed", self.rebuild_status_icon_menu)
@@ -1258,13 +1539,29 @@ class WarpApplication(Gtk.Application):
 
     def _remote_ops_changed(self, local_machine, name):
         self.window.refresh_remote_machine_view()
-        self.update_inhibitor_state(local_machine)
+
+        active_ops = self.server.get_active_op_count(incoming_only=True) if self.server else 0
+        self.update_inhibitor_state()
+
+        if active_ops > 0:
+            util.free_space_monitor.start()
+        else:
+            util.free_space_monitor.pause()
+
+        if self.app_restarting:
+            self.window.update_restart_dialog_status(active_ops)
+
+    def _manual_connect_result(self, _, initiated_here, result, msg):
+        if not initiated_here:
+            return
+        if self.window.connect_dialog:
+            self.window.connect_dialog.on_connection_result(result, msg)
 
     def add_simulated_widgets(self):
         import testing
         testing.add_simulated_widgets(self)
 
-    def update_inhibitor_state(self, local_machine):
+    def update_inhibitor_state(self):
         any_active_ops = False
 
         remotes = self.server.list_remote_machines()
@@ -1294,20 +1591,20 @@ class WarpApplication(Gtk.Application):
 
     def update_status_icon_from_preferences(self):
         if prefs.use_tray_icon():
-            if self.status_icon == None:
-                self.status_icon = XApp.StatusIcon()
+            if self.status_icon is None:
+                self.status_icon = XApp.StatusIcon(name="org.x.StatusIcon.warpinator")
                 self.status_icon.connect("activate", self.on_tray_icon_activate)
-                self.update_status_icon_online_state(self.server != None)
+                self.update_status_icon_online_state(self.server is not None)
                 self.hold()
             self.rebuild_status_icon_menu()
         else:
-            if self.status_icon != None:
+            if self.status_icon is not None:
                 self.status_icon.set_visible(False)
                 self.status_icon = None
                 self.release()
 
     def update_status_icon_online_state(self, online=True):
-        if self.status_icon == None:
+        if self.status_icon is None:
             return
 
         if online:
@@ -1320,12 +1617,12 @@ class WarpApplication(Gtk.Application):
         self.rebuild_status_icon_menu()
 
     def rebuild_status_icon_menu(self, remote_machine=None):
-        if self.status_icon == None:
+        if self.status_icon is None:
             return
 
         menu = Gtk.Menu()
 
-        if self.server != None:
+        if self.server is not None:
             self.add_favorite_entries(menu)
 
         item = Gtk.MenuItem(label=_("Open save folder"))
@@ -1439,23 +1736,29 @@ class WarpApplication(Gtk.Application):
     def on_tray_icon_activate(self, icon, button, time):
         self.window.toggle_visibility(time)
 
-def main(test=False, debug=False):
+
+def main():
     import signal
 
-    log_handler = logging.StreamHandler(sys.stdout)
-    log_handler.setFormatter(util.WarpLogFormatter())
-    logging.root.addHandler(log_handler)
+    try:
+        config.sandbox_mode = os.environ["WARPINATOR_SANDBOX_MODE"]
+    except KeyError as e:
+        pass
 
-    if debug:
-        logging.root.setLevel(logging.DEBUG)
-    else:
-        logging.root.setLevel(logging.INFO)
+    try:
+        w = WarpApplication(testing=False)
+        signal.signal(signal.SIGINT, lambda s, f: w.exit_warp())
+        signal.signal(signal.SIGTERM, lambda s, f: w.exit_warp())
 
-    w = WarpApplication(test)
-    signal.signal(signal.SIGINT, lambda s, f: w.exit_warp())
-    signal.signal(signal.SIGTERM, lambda s, f: w.exit_warp())
+        ret = w.run(sys.argv)
 
-    return w.run(sys.argv)
+        if w.app_restarting:
+            ret = misc.EXIT_CODE_RESTART_BWRAP
+    except Exception as e:
+        print(e)
+        ret = 1
+
+    return ret
 
 if __name__ == "__main__":
-    main()
+    sys.exit(main())
